/*
	This code is part of the MCJoin project.
	Authored by Steven Begley (sbegley@latrobe.edu.au) as part of my PhD candidature.
	La Trobe University,
	Melbourne, Australia.
*/
#pragma once
#include "SBUN.h"
#include "CRelation.h"

class CFlipFlop
{
public:
	CFlipFlop(): _input(NULL), _output(NULL), _cardinality(0), _histogram(NULL), _numberOfPartitions(0)
	{
	}

	~CFlipFlop()
	{
	}

	__forceinline void setCardinality(uint32 cardinality)
	{
		_cardinality = cardinality;
	}

	__forceinline void setInput(CRelation* input)
	{
		_input = input;
	}

	__forceinline void setOutput(CRelation* output)
	{
		_output = output;
	}

	__forceinline void setRelations(const CRelation* input, CRelation* output, uint32 cardinality)
	{
		_input = input;
		_output = output;

		setCardinality(cardinality);
	}

	void go(uint32 D, uint32 R)
	{
		buildHistogram(D, R);
		buildPrefixSum();
		scatter(D, R);
		dropHistogram();
	}

	void go(uint32 D, uint32 R, uint32 start, uint32 end, uint32 baseIndex, bool retainPayload)
	{	
		buildHistogram(D, R, start, end);
		buildPrefixSum();
		scatter(D, R, start, end, baseIndex, retainPayload);
		dropHistogram();
	}

private:

	__forceinline void dropHistogram()
	{
		if(_histogram != NULL)
		{
			free(_histogram);
			_histogram = NULL;
		}
	}

	__forceinline void buildHistogram(uint32 D, uint32 R)
	{
		_numberOfPartitions = 1 << D;

		_histogram = (uint64*)calloc(_numberOfPartitions, sizeof(uint64));

		register uint32 M = ( ( 1 << D ) - 1) << R;
		register SBUN* p = _input->getBATPointer();
		register uint32 idx(0);

		for(register uint32 index = 0; index < _cardinality; ++index)
		{
			idx = (p->key) & M;
			idx >>= R;

			++(_histogram[idx]);

			++p;
		}
	}

	__forceinline void buildHistogram(uint32 D, uint32 R, uint32 start, uint32 end)
	{	
		_numberOfPartitions = 1 << D;	
		_histogram = (uint64*)calloc(_numberOfPartitions, sizeof(uint64));

		register uint32 M = ( ( 1 << D ) - 1) << R;
		register SBUN* p = &_input->getBATPointer()[start];

		register uint32 idx(0);
		register uint32 itemCount(end - start);

		for(register uint32 index = 0; index < itemCount; ++index)
		{
			idx = (p->key) & M;
			idx >>= R;

			++(_histogram[idx]);

			++p;
		}
	}

	__forceinline void buildPrefixSum()
	{
		uint64* h = _histogram;
		uint64 thisValue(0);
		register uint64 sum(0);
		
		for(uint32 index = 0; index < _numberOfPartitions; ++index )
		{						
			thisValue = *h;

			*h = sum;

			sum += thisValue;
			++h;	
		}

		uint64 histValue(0);
		SBUN* p = _output->getBATPointer();
		for(uint32 index = 0; index < _numberOfPartitions; ++index )
		{		
			histValue = _histogram[index];
			_histogram[index] = (uint64)(p+histValue);
		}
	}

	__forceinline void scatter(uint32 D, uint32 R)
	{
		register uint32 M = ( ( 1 << D ) - 1) << R;
		register SBUN* p = _input->getBATPointer();		
		register SBUN* write(NULL);
		register uint32 idx(0);

		for(register uint32 index = 0; index < _cardinality; ++index)
		{
			idx = (p->key) & M;
			idx >>= R;

			write = (SBUN*)_histogram[idx];
			*write = *p;
			_histogram[idx] += sizeof(SBUN);

			++p;
		}
	}

	__forceinline void scatter(uint32 D, uint32 R, uint32 start, uint32 end, uint32 baseIndex, bool retainPayload)
	{
		register uint32 M = ( ( 1 << D ) - 1) << R;
		register SBUN* p = &_input->getBATPointer()[start];		
		register SBUN* write(NULL);
		register uint32 idx(0);

		register uint32 itemCount(end - start);

		for(register uint32 index = 0; index < itemCount; ++index)
		{
			idx = (p->key) & M;
			idx >>= R;

			write = (SBUN*)_histogram[idx];
			//*write = *p;
			write->key = p->key;
			write->payload = retainPayload?p->payload:((start + index) - baseIndex);

			_histogram[idx] += sizeof(SBUN);

			++p;
		}
	}


	void showDebug()
	{
		for(uint32 counter = 0; counter < _numberOfPartitions; ++counter)
		{
			//TRACE2("Partition %d : [%d]\n", counter, _histogram[counter]);
			cout << "Partition " << counter << " : [" << _histogram[counter] << "]" << endl;
		}
	}


private:
	uint32 _numberOfPartitions;
	uint32 _cardinality;
	uint64* _histogram;
	const CRelation* _input;
	CRelation* _output;
};