#include <iostream>
#include <vector>
#include "DataTypes.h"
#include "CRelation.h"

using namespace std;

struct SOptions
{
	uint32 output;				// 0 = text file, 1 = binary, 2 = both
	uint32 cardinality;	
	double selectivity;			// This adjusts the upper range of the random numbers as a ratio of the relation size. Typically, this would be set to 1 for
								// an R relation, and e.g. 0.1 for an S relation which means any tuple in R has a 10% chance of finding a match in S.
								// For a relation of 1 million cardinality, setting the selectivity to 0.1 would mean the upper bound for random numbers would be 10 million;
								// setting it to 0.5 would set the upper bound to 2 million, and so forth.
	double skew;
	double mean;
	string relationFilename;
	bool autoGenerateFilename;
	bool useAlternateGenerator;

	
	SOptions() : output(2), cardinality(1000000), selectivity(1), skew(0), mean(0.5), autoGenerateFilename(false), useAlternateGenerator(false)
	{
		relationFilename = "";		
	}
};

inline bool file_exists(const string &name) 
{
	ifstream f(name.c_str());

	if (f.good()) {
		f.close();
		return true;
	} else {
		f.close();
		return false;
	}   
}

void showUsage()
{
	cout << "Usage: mkrel filename [options]" << endl;
	cout << "[options] consists of:" << endl;
	cout << "    -o output (0 = text, 1 = binary, 2 = both) (e.g. -o2)" << endl;
	cout << "    -c cardinality (e.g. -c1000000)" << endl;
	cout << "    -s selectivity (e.g. -s0.1)" << endl;
	cout << "    -g skew data set (e.g. -g0.01, if not set (or -g0 then distribution is uniform))" << endl;
	cout << "    -m mean for skew (0..1) (e.g. -m0.5, data is skewed around the center point of the random number range)" << endl;
	cout << "	 -a automatically generate filename (needs a hint if it is R or S)" << endl;
	cout << "	 -z use alternate generator seed" << endl;
	cout << endl;
	cout << "Default settings are: -o2 -c1000000 -s1 -g0 -m0.5" << endl;
}

string toUpper(string& str)
{
	std::string converted;

	for(short i = 0; i < str.size(); ++i)
		converted += toupper(str[i]);

	return converted;
}

void generateFileName(SOptions *options)
{
	ostringstream ss;

	ss.str("");
	ss << toUpper(options->relationFilename);
	
	if (options->skew == 0)
	{
		ss << "_U";
	}
	else
	{
		ss << "_S" << (options->skew * 100);
	}

	if(options->selectivity != 1.0)
	{
		ss << "_" << (options->selectivity * 100) << "P";
	}

	ss << "_" << (options->cardinality / 1000000) << "M";

	options->relationFilename = ss.str();
}

bool buildRelations(SOptions *options)
{
	bool result(true);
	CRelation *relation;
	
	uint32 upperBound = (uint32)(options->cardinality / options->selectivity);

	try
	{
		if(options->autoGenerateFilename == true)
		{
			generateFileName(options);
		}

		relation = new CRelation(options->cardinality);
		if(options->skew == 0)
		{
			relation->initialiseSet_Uniform(1, upperBound, options->useAlternateGenerator);
		}
		else
		{
			relation->initialiseSet_Gaussian(1, upperBound, (upperBound / 2), options->skew, options->useAlternateGenerator);
		}		
		if(options->output == 0 || options->output == 2)
		{
			relation->save((options->relationFilename + ".txt").c_str(), false);
		}
		if(options->output == 1 || options->output == 2)
		{
			relation->save((options->relationFilename + ".dat").c_str(), true);
		}
		delete relation;
		relation = NULL;				
	}
	catch (...)
	{
		result = false;
	}

	return result;
}

bool parseArguments(vector<string> &arguments, SOptions *options)
{
	bool returnValue(true);
	string buffer;	
	uint32 untaggedParam(0);

	if(arguments.size() == 1)
	{
		showUsage();
	}
	else
	{
		for(string& s : arguments)
		{
			buffer.clear();
			
			if (s[0] == '-')
			{		
				buffer += s.substr(2);	
				
				switch (s[1])
				{
				case 'c':										
					options->cardinality = atoi(buffer.c_str());
					break;
				case 's':					
					options->selectivity = atof(buffer.c_str());
					break;
				case 'o':					
					options->output = atoi(buffer.c_str());
					break;
				case 'g':
					options->skew = atof(buffer.c_str());
					break;
				case 'm':
					options->mean = atof(buffer.c_str());
					break;
				case 'a':
					options->autoGenerateFilename = true;
					break;
				case 'z':
					options->useAlternateGenerator = true;
					break;
				default:
					cout << "unknown option : -"  << s[1]<< endl;
					returnValue = false;
				}
			}
			else
			{				
				switch (untaggedParam)
				{
				case 0:
					// This will be the executable name. Do not want.
					break;
				case 1:
					// This will be the relation filename.
					options->relationFilename = s;
					break;
				default:
					// Just going to ignore anything else.
					break;
				}

				++untaggedParam;
			}
		}
	}

	return returnValue;
}

int main(int argc, char* argv[])
{
	int returnValue(0);
	SOptions options;

	vector<string> arguments(argv, argv + argc);
	if(parseArguments(arguments, &options))
	{
		cout << "Executing...";

		if(buildRelations(&options))
		{
			cout << "done!" << endl;
		}
		else
		{
			cout << "failed :(" << endl;
		}		
	}
	else
	{		
		returnValue = -1;
	}

	return returnValue;
}

